import MDAnalysis
from MDAnalysis.coordinates.base import Timestep
import numpy as np

class NUMPYReader(MDAnalysis.coordinates.base.Reader):
    """
    Reads coordinates from a (num_frames, num_atoms x 3) numpy array
    into MDAnalysis universes. Modified from the XYZ reader in the MDAnalysis distribution
    As it is a numpy array, all in memory.
    """
    format = "Numpy"
    units = {'time': 'ns', 'length': 'Angstrom'}

    def __init__(self, filename='inmemory.numpy', **kwargs):

        self.nparray = kwargs['nparray']
        self.nparray = self.nparray.reshape(self.nparray.shape[0], -1)
        
        self.fixed = 0
        self.skip = 1
        self.periodic = False
        self.skip_timestep = 0
        self.delta = 0
        self.n_atoms = self.nparray[0].shape[0] / 3
        self.n_frames = self.nparray.shape[0]
        self.ts = Timestep(self.n_atoms)
        self.ts.frame = -1
        # Read in the first timestep
        self._read_next_timestep()

    def __iter__(self):
        self.ts.frame = -1 
        while True:
            try:
                yield self._read_next_timestep()
            except EOFError:
                self.close()
                raise StopIteration

    def _read_next_timestep(self, ts=None):
        # check that the timestep object exists
        if ts is None:
            ts = self.ts

        if ts.frame < self.n_frames - 1:
            # stop when the cursor has reached the end of that block
            ts.frame += 1
            ts._unitcell = np.zeros(6, np.float32)
            ts._x[:] = self.nparray[ts.frame][0::3]
            ts._y[:] = self.nparray[ts.frame][1::3]
            ts._z[:] = self.nparray[ts.frame][2::3]
            return ts
        raise EOFError

    def _jump_to_frame(self,frame):
        ts = self.ts
        ts._unitcell = np.zeros(6, np.float32)
        ts._x[:] = self.nparray[frame][0::3]
        ts._y[:] = self.nparray[frame][1::3]
        ts._z[:] = self.nparray[frame][2::3]
        ts.frame = frame
        return ts

    
    def populate_ts(self, ts, frame):
        ts._unitcell = np.zeros(6, np.float32)
        ts._x[:] = self.nparray[frame][0::3]
        ts._y[:] = self.nparray[frame][1::3]
        ts._z[:] = self.nparray[frame][2::3]
        ts.frame = frame 

    def rewind(self):
        """reposition on first frame"""
        # reset ts
        ts = self.ts
        ts.status = 1
        ts.frame = -1
        ts.step = 0
        ts.time = 0
        self.skip_timestep = 0
        # the next method is inherited from the Reader Class and calls _read_next_timestep
        self.next()

    def Writer(self, filename, numatoms=None, **kwargs):
        #NB. Just copied from DCD.py
        #numatoms = kwargs.pop('numatoms', self.numatoms)
        kwargs.setdefault('start', 0)
        kwargs.setdefault('step', self.skip)
        kwargs.setdefault('delta', 10)
        kwargs.setdefault('remarks', 'All a bit too much hardcoded')
        from MDAnalysis.coordinates.DCD import DCDWriter
        return DCDWriter(filename, numatoms, **kwargs)

    def __getitem__(self, frame):
        if (np.dtype(type(frame)) != np.dtype(int)) and (type(frame) != slice):
            raise TypeError
        if np.dtype(type(frame)) == np.dtype(int):
            if frame < 0:
                # Interpret similar to a sequence
                frame += len(self)
            if (frame < 0) or (frame >= len(self)):
                raise IndexError            
            return self._jump_to_frame(frame)
        elif type(frame) == slice: # if frame is a slice object
            if not ((type(frame.start) == int or frame.start is None) and
                    (type(frame.stop) == int or frame.stop is None) and
                    (type(frame.step) == int or frame.step is None)):
                raise TypeError("Slice indices are not integers")
            def iterDCD(start=frame.start, stop=frame.stop, step=frame.step):
                start, stop, step = self._check_slice_indices(start, stop, step)
                for i in xrange(start, stop, step):
                    yield self[i]
            return iterDCD()



#!/usr/bin/env python
'''
The pcz class.
'''
from MDAnalysis.analysis.align import *
import numpy as np
from scipy.linalg import *
import struct
import logging as log
import sys
from time import time
import warnings

warnings.simplefilter("ignore")

try:
    import h5py
    h5py_available = True
except ImportError:
    h5py_available = False

class Pcz:
    def __init__(self, cofasu, version='PCZ6',target=None, covar=None, quality=90.0,
                 req_evecs=None, rank = 0, preload=False, fastmethod=False):
        '''
        Initialises a new pcz object with the data from the given
        cofasu object:
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> c = cofasu.Cofasu(cofasu.Fasu(topfile, trjfile, filter='name CA and resid 1-10'))
        >>> p = Pcz(c)

        
        target can be a precalculated global average
        structure:

        >>> target = c.fitted_average()-10.0
        >>> p = Pcz(c, target=target)

        covar can be a precalculated covariance matrix, in which case the
        corresponding target structure must also be given:
        >>> cm = c.cov()
        >>> p = Pcz(c, target=target, covar=cm)

        >>> p = Pcz(c, covar=cm)
        Traceback (most recent call last):
           ...
        ValueError: A defined covariance matrix requires a defined target.


        The quality setting defaults to 90%:
        >>> p = Pcz(c, quality=95)
        >>> p = Pcz(c, quality=120)
        Traceback (most recent call last):
           ...
        ValueError: quality must lie in the range 0:100.

        If preload==True, then an attempt will be made to load all the
        trajectory data into memory. This means slower initialisation but
        faster access to data later.

        >>> c = cofasu.Cofasu(cofasu.Fasu(topfile, trjfile, filter='name CA and resid 1-10'))
        >>> p = Pcz(c)
        >>> ev1 = p.evals()
        >>> p2 = Pcz(c, preload=True)
        >>> ev2 = p2.evals()
        >>> print(np.allclose(ev1, ev2))
        True
 
        If fastmethod=True then a fast approximate diagonalisation
        method is used.

        >>> f = cofasu.Fasu(topfile, trjfile, filter='resid 1-30')
        >>> c = cofasu.Cofasu(f)
        >>> p1 = Pcz(c)
        >>> ev1 = p1.evals()
        >>> p2 = Pcz(c, fastmethod=True)
        >>> ev2 = p2.evals()
        >>> print(np.allclose(ev1, ev2))
        True

        '''

        self.preloaded = preload
        self.version = version
        self.cofasu = cofasu
        self.quality = quality
        self.rank = rank
        self.natoms = self.cofasu.natoms
        self.nframes = self.cofasu.numframes()

        if quality < 0 or quality > 100:
            raise ValueError('quality must lie in the range 0:100.')

        if rank == 0:
            log.info('Pcz: {0} atoms and {1} snapshots'.format(self.natoms, self.nframes))
        if covar is None:
            if rank == 0:
                log.info('Pcz: least-squares fitting snapshots')
            if target is None:
                time_avg_0 = time()
                self._avg = cofasu.fitted_average()
                time_avg_1 = time()
                if rank == 0:
                    log.info(
                'Pcz: Time for trajectory fitting: {0:.2f} s\n'.format(time_avg_1 - time_avg_0))

            else:
                self._avg = target

            if rank == 0:
                log.info('Pcz: calculating covariance matrix')
            if fastmethod:
                # adapted from Ian Dryden's R code. If you have
                # n atoms and p snapshots, then the conventional
                # way to do the pca is to calculate the [3n,3n]
                # covariance matrix and then diagonalise that.
                # However if p < 3n, then the last 3n-p eigenvectors
                # and values are meaningless anyway, and instead
                # you can calculate the [p,p] covariance matrix,
                # diagonalise that, and then do some
                # data massaging to recover the full eigenvectors
                # and eigenvalues from that. Here we extend the
                # approach to situations where 3n is just too big
                # for diagonalisation in a reasonable amount of
                # time, by just taking a selection of snapshots
                # from the full set (<< 3n) and applying this
                # approach. Obviously this is an approximate
                # method, but it may be good enough.
                if rank == 0:
                    log.info("Using fast approximate diagonalisation method")
                nsamples = min(100,self.nframes)
                tmptrj = np.zeros((3*self.natoms,nsamples))
                stepsize = self.nframes/nsamples
                j = 0
                for i in range(nsamples):
                    x2 = self.cofasu.coords(j,target=self._avg) - self._avg
                    tmptrj[:,i] = x2.flatten()
                    j += stepsize
                cv = np.dot(tmptrj.T,tmptrj.conj())/nsamples
            else:
                cv = self.cofasu.cov(self._avg, preload=preload)
        else:
            # Covariance matrix supplied. This requires that the corresponding
            # target structure is given as well.
            if target is None:
                raise ValueError('A defined covariance matrix requires a defined target.')
            else:
                self._avg = target
                cv = covar

        if rank == 0:
            log.info('Pcz: diagonalizing covariance matrix')
        time_diag_cov_0 = time()
        w, v = eigh(cv)
        if fastmethod:
            vv = np.zeros(nsamples)
            z = np.dot(tmptrj,v)
            for i in range(nsamples):
                vv[i] = np.sqrt((z[:,i]*z[:,i]).sum())
                z[:,i] = z[:,i]/vv[i]

            w2 = np.sqrt(abs(w/nsamples))*vv
            w = w2[w2.argsort()]
            v = z[:,w2.argsort()]

        cs = np.cumsum(w[::-1])
        self.totvar = cs[-1]
        tval = cs[-1] * self.quality / 100
        i = 0
        while cs[i] < tval:
            i += 1

        i += 1
        self.nvecs = i
        # override this with req_evecs, if given:
        if req_evecs is not None:
            if req_evecs > len(w):
                if rank == 0:
                    log.error('Pcz: you asked for {0} eigenvectors but there are only {1} available.'.format(req_evecs, len(w)))
            else:
                self.nvecs = req_evecs
                i = req_evecs

        self._evals = w[-1:-(i + 1):-1]
        self._evecs = v[:, -1:-(i + 1):-1].T
        time_diag_cov_1 = time()
        if rank == 0:
            log.info(
                'Pcz: Time for diagonalizing covariance matrix: {0:.2f} s\n'.format(time_diag_cov_1 - time_diag_cov_0))

        if rank == 0:
            log.info('Pcz: calculating projections')
        time_proj_calc_0 = time()
        if self.preloaded:
            trj = np.zeros((self.nframes, self.natoms * 3))
            for i in range(self.nframes):
                mob = self.cofasu.coords(i)
                mob -= mob.mean(axis=0)
                R, rms = rotation_matrix(self._avg, mob)
                trj[i, :] = np.dot(mob, R).flatten() - self._avg.flatten()

            self._projs = np.dot(trj, self._evecs.T).T
        else:
            if rank == 0:
                log.info("Calculating projections big files method")
            max_ts = 1000
            num_blocks = self.nframes/max_ts
            ts_extra = self.nframes%max_ts
            self._projs = np.zeros((self.nvecs, self.nframes))
            trj = np.zeros((max_ts, self.natoms * 3))
            for i in range(num_blocks):
                for j in range(max_ts):
                    index = i*max_ts + j
                    mob = self.cofasu.coords(index)
                    mob -= mob.mean(axis=0)
                    R, rms = rotation_matrix(self._avg, mob)
                    trj[j, :] = np.dot(mob, R).flatten() - self._avg.flatten()

                self._projs[:,i*max_ts:(i+1)*max_ts] = np.dot(trj, self._evecs.T).T
            trj = np.zeros((ts_extra, self.natoms * 3))
            for j in range(ts_extra):
                index = num_blocks*max_ts + j
                mob = self.cofasu.coords(index)
                mob -= mob.mean(axis=0)
                R, rms = rotation_matrix(self._avg, mob)
                trj[j, :] = np.dot(mob, R).flatten() - self._avg.flatten()

            self._projs[:,num_blocks*max_ts:] = np.dot(trj, self._evecs.T).T
        time_proj_calc_1 = time()
        if rank == 0:
            log.info('Pcz: Time for calculating projections: {0:.2f} s\n'.format(time_proj_calc_1 - time_proj_calc_0))

    def numframes(self):
        """
        Method to match the cofasu equivalent
        """
        return self.nframes

    def avg(self):
        """
        Returns the average structure contained in the pcz file
        as an (natoms,3) numpy array.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> print(np.allclose(p.avg()[0], np.array([ 31.323149, 61.575380, 40.136298]), atol=0.001, rtol=0.001))
        True

        """
        return self._avg

    def eval(self, ival):
        """
        Returns an eigenvalue from the file.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> print(np.allclose(p.eval(4), np.array([2.6183940]), rtol=0.001, atol=0.001))
        True

        """
        if ival >= self.nvecs:
            print 'Error - only ', self.nvecs, ' eigenvectors present'
            return 0.0
        else:
            return self._evals[ival]

    def evals(self):
        """
        Returns an array of all eigenvalues in the file.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> print(np.allclose(p.evals()[4], np.array([2.6183940]), rtol=0.001, atol=0.001))
        True

        """
        return self._evals

    def evec(self, ivec):
        """
        Returns a chosen eigenvector from the file in the
        form of a (3*natoms) numpy array.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> print(np.allclose(abs(p.evec(1)[12]), np.array([0.00865751377]), rtol=0.001, atol=0.001))
        True

        """
        if ivec >= self.nvecs:
            print 'Error - only ', self.nvecs, 'eigenvectors present'
            return None
        else:
            return self._evecs[ivec, :]

    def evecs(self):
        """
        Returns all eigenvectors in the file in the form of a
        (nvecs,3*natoms) numpy array.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> e = p.evecs()
        >>> print(e.shape)
        (18, 471)

        >>> element = abs(e[1,12])
        >>> print(np.allclose(element, np.array([0.0086575138]), rtol=0.001, atol=0.001))
        True

        """
        return self._evecs

    def proj(self, iproj):
        """
        Returns an array of the projections along a given eigenvector. There
        will be one value per snapshot.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> prj = abs(p.proj(3))
        >>> print(np.allclose(prj[21], 0.33430696, rtol=0.001, atol=0.001))
        True

        """
        if iproj >= self.nvecs:
            print 'Error - only ', self.nvecs, 'eigenvectors present'
            return None
        else:
            return self._projs[iproj, :]

    def scores(self, framenumber):
        """
        Method that returns the scores (projections) corresponding to
        a chosen snapshot (zero-based).
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> s = abs(p.scores(12))
        >>> print(np.allclose(s[3], 0.58309314, rtol=0.001, atol=0.001))
        True

        """
        if( framenumber >= self.nframes):
             return None
        else:
             x = np.zeros(self.nvecs)
             for i in range(self.nvecs):
                 x[i] = self.proj(i)[framenumber]
             return x

    def coords(self, framenumber):
        """
        Synonym for frame() method, to match cofasu.
        """
        return self.frame(framenumber)

    def frame(self,framenumber):
        """
        Method to return the coordinates of the given frame - i.e.
        to decompress a snapshot. The data is returned as a (natoms,3) 
        numpy array.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c, quality=95)
        >>> ref = c.coords(5)
        >>> x = p.frame(5)
        >>> cofasu.rmsd(ref, x) < 0.19
        True

        """
        if(framenumber >= self.nframes):
            return None
        else:
            scores = self.scores(framenumber)
            return self.unmap(scores)


    def closest(self, scores):
        """
        Method to find the index of the frame with scores closest to the
        target values.
        """
        ns = len(scores)
        if self.preloaded:
            temp = self._projs
        else:
            temp = np.zeros((self.nvecs, self.nframes))
            for vec in range(self.nvecs):
                temp[vec, :] = self.proj(vec)

        best = 0
        err = ((temp[0:ns, 0] - scores) * (temp[0:ns, 0] - scores)).sum()
        for frame in range(self.nframes):
            newerr = ((temp[0:ns, frame] - scores) * (temp[0:ns, frame] - scores)).sum()
            if newerr < err:
                err = newerr
                best = frame
        return best

    def unmap(self,scores):
        """
        Method to return the coordinates corresponding to a given
        set of scores. If the scores vector has less than nvec elements,
        Pczfile.expand is called to fill in the missing values.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> a = p.avg()
        >>> a2 = p.unmap(np.zeros(p.nvecs))
        >>> print(cofasu.rmsd(a, a2) < 0.001)
        True

        """
        x = self.avg()
        if len(scores) < self.nvecs:
            scores = self.expand(scores)
        for i in range(self.nvecs):
            x = x + (self.evec(i)*scores[i]).reshape((self.natoms,3))
        return x

    def expand(self, scores):
        """
        Method to complete a truncated list of scores with values for
        missing eigenvectors. Basically the pcz file is scanned to
        find the frame with the best match to the scores given, the
        missing values are then copied directly from this frame.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> e = abs(p.expand([1.0, 2.0, 3.0]))
        >>> print(np.allclose(e[3:6], [2.5556779, 0.9558872, 2.2131005], rtol=0.001, atol=0.001))
        True

        """
        nvecs = self.nvecs
        ns = len(scores)
        if ns >= nvecs:
            return scores
        else:
            newscores = np.zeros(nvecs)
            newscores[0:ns] = scores
            temp = self._projs
            best = 0
            err = ((temp[0:ns,0]-scores)*(temp[0:ns,0]-scores)).sum()
            newscores[ns:nvecs] = temp[ns:nvecs,0]
            for frame in range(self.nframes):
                newerr = ((temp[0:ns,frame]-scores)*(temp[0:ns,frame]-scores)).sum()
                if newerr < err:
                    err = newerr
                    newscores[ns:nvecs] = temp[ns:nvecs,frame]
            return newscores

    def map(self,crds):
        """
        Method to map an arbitrary coordinate set onto the PC model. The
        coordinate set should be a (natom,3) array-like object that matches
        (for size) what's in the pczfile. An array of projections will be 
        returned, one value for each eignevector in the pcz file.
        >>> topfile = "../../../../test/2ozq.pdb"
        >>> trjfile = "../../../../test/2ozq.dcd"
        >>> from MDPlus.core import cofasu
        >>> f = cofasu.Fasu(topfile, trjfile, filter='name CA')
        >>> c = cofasu.Cofasu(f)
        >>> p = Pcz(c)
        >>> crds = (c.coords(10) + c.coords(20))* 0.5
        >>> print(np.allclose(abs(p.map(crds)[:3]),[0.3844803, 1.6537842, 2.29433996], rtol=0.001, atol=0.001))
        True

        """
        crdset=np.array(crds)
        if np.shape(crdset) == (self.natoms, 3):
          avg = np.reshape(self.avg(),(self.natoms,3))
          rot = np.zeros(9, dtype=np.float64)
          crdset_cog = crdset.sum(axis=0)/crdset.shape[0]
          avg_cog = avg.sum(axis=0)/avg.shape[0]
          c = crdset.copy()
          c -= crdset_cog
          rmsd = qcp.CalcRMSDRotationalMatrix((avg-avg_cog).T.astype(np.float64),
                                    c.T.astype(np.float64),
                                    crdset.shape[0], rot, None)
          R = np.matrix(rot.reshape(3,3))
          c = c * R
          c += avg_cog
          c -= avg
          prj = np.zeros(self.nvecs)
          for i in range(self.nvecs):
              prj[i]=(np.dot(c.flatten(),self.evec(i)))
          return prj
        else:
          return None
  

    def write(self, filename,  title='Created by pcz.write()'):
        """
        Write out the PCZ file. At the moment only the PCZ4 and PCZ6 formats
        are  implemented.
        """
        if self.version == 'PCZ7' and not h5py_available:
            log.info("WARNING: The PCZ6 format will be used because the h5py module required for PCZ7 is not available. Please install the PCZ7 extra requirements to be able to use it. The command to do this is: pip install pyPcazip[PCZ7]")
            self.version = 'PCZ6'

        if self.version == 'UNKN':
            if h5py_available:
                self.version = 'PCZ7'
            else:
                self.version = 'PCZ6'

        if self.version != 'PCZ4' and self.version != 'PCZ6' and self.version != 'PCZ7':
            raise TypeError('Only PCZ4/6/7 formats supported')

        if self.rank == 0:
            log.info("Using "+self.version+" format")

        if self.version == 'PCZ4' or self.version == 'PCZ6':
            f = open(filename, 'wb')
            f.write(struct.pack('4s80s3if', self.version, title, self.natoms, self.nframes, self.nvecs, self.totvar))
            f.write(struct.pack('4i', 0, 0, 0, 0))
            for v in self.avg().flatten():
                f.write(struct.pack('f', v))
            for i in range(self.nvecs):
                for v in self.evec(i):
                    f.write(struct.pack('f', v))
                f.write(struct.pack('f', self.eval(i)))

                # All in memory
                if self.preloaded:
                    projection = self.proj(i)

                # loop file (low memory demand)
                else:
                    projection = np.zeros(self.nframes)
                    for t in range(self.nframes):

                        mob = self.cofasu.coords(t)
                        mob -= mob.mean(axis=0)
                        R, rms = rotation_matrix(self._avg, mob)

                        time_step_coords = np.dot(mob, R).flatten() - self._avg.flatten()
                        projection[t] = np.dot(time_step_coords, self._evecs[i].T).T

                if self.version == 'PCZ4':
                    for v in projection:
                        f.write(struct.pack('f', v))
                elif self.version == 'PCZ6':
                    pinc = (projection.max() - projection.min()) / 65534
                    p0 = (projection.max() + projection.min()) / 2
                    f.write(struct.pack('2f', p0, pinc))
                    for v in projection:
                        f.write(struct.pack('h', np.int16((v - p0) / pinc)))

                else:
                    print 'Error - only PCZ4 and PCZ6 formats supported'

            f.close()
            return
        elif self.version == 'PCZ7':
            if not h5py_available:
                print "Error: h5py module not available. Please install the PCZ7 extra requirements to be able to use it. The command to do this is: pip install pyPcazip[PCZ7]"
                sys.exit(0)
            f = h5py.File(filename, "w")
            # Write evecs and evalues
            evec_array = []
            eval_array = []
            for evec_index in xrange(self.nvecs):
                evec_array.append([])
                for v in self.evec(evec_index):
                    evec_array[evec_index].append(v)
                eval_array.append(self.eval(evec_index))
            f.create_dataset("evec_dataset", (self.nvecs, (3 * self.natoms)), dtype='f', data=np.array(evec_array))
            f.create_dataset("eval_dataset", (self.nvecs,), dtype='f', data=np.array(eval_array))
            # Write reference coordinates
            f.create_dataset("ref_coord_dataset", (len(self.avg().flatten()),), dtype='f',
                             data=np.array(self.avg().flatten()))
            # Write properties
            f.attrs['version'] = self.version
            f.attrs['title'] = title
            f.attrs['natoms'] = self.natoms
            f.attrs['nframes'] = self.nframes
            f.attrs['nvecs'] = self.nvecs
            f.attrs['quality'] = self.totvar

            # Loop on every ts
            proj_dataset = f.create_dataset("proj_dataset", (self.nframes,
            self.nvecs), dtype='int16')
            p0_dataset = f.create_dataset("p0_dataset", (self.nframes,), dtype='f')
            pinc_dataset = f.create_dataset("pinc_dataset", (self.nframes,), dtype='f')



            for ts_index in xrange(self.nframes):
                projection_values = []
                for evec_index in xrange(self.nvecs):
                    # Prepare coords of ts
                    if self.preloaded:
                        projection_values.append(self.proj(evec_index)[ts_index])
                    else:
                        mob = self.cofasu.coords(ts_index)
                        mob -= mob.mean(axis=0)
                        R, rms = rotation_matrix(self._avg, mob)
                        time_step_coords = np.dot(mob, R).flatten() - self._avg.flatten()
                        projection_values.append((np.dot(time_step_coords, self._evecs[evec_index].T).T).item(0))

                # Write coords of ts
                projection_values = np.array(projection_values)
                pinc = (projection_values.max() - projection_values.min()) / 65534
                if pinc == 0:
                    pinc == numpy.nextafter(0, 1)
                p0 = (projection_values.min() + projection_values.max()) / 2
                p0_dataset[ts_index] = p0
                pinc_dataset[ts_index] = pinc
                projection_values = projection_values - p0
                proj_dataset[ts_index] = (projection_values / pinc).astype(np.int16)


        else:
            raise TypeError('Only PCZ4/6/7 formats supported')

if __name__ == "__main__":
    import __builtin__
    __builtin__.parallelbuiltin = False
    import doctest
    doctest.testmod()



def pdb2selection(pdbfile):
    '''
    A little utility function to convert 'mask' pdb files into
    MDAnalysis-style selection strings. Basically it reads the second column
    (atom number) and uses it to construct 'bynum' selections. Runs
    of consecutive numbers are expressed in start:stop form.
    '''

    sel = ''
    i = 0
    j = 0
    with open(pdbfile, 'r') as f:
        for line in f:
            if line.find('ATOM') == 0 or line.find('HETATM') == 0:
                k = int(line.split()[1])
                # the next line catches the initialization process:
                if i == 0:
                    i = k
                    j = k - 1
                # are we in a run of consecutive numbers?:
                if k == j + 1:
                    j = k
                else:
                    # time to write out another selection:
                    sel = sel + ' bynum {0}:{1} or'.format(i, j)
                    i = k
                    j = k
                # end-of-file reached. Make sure last selection is included:
    if i > 0 and j > 0:
        sel = sel + ' bynum {0}:{1} or'.format(i, j)
    if len(sel) > 3:
        # remove the trailing ' or':
        sel = sel[:-3]
    return sel




# The mapping module. Creates and does various useful things with
# multidimensional histograms.

import numpy as np
import scipy.ndimage as nd

class Map():
    
    def __init__(self, arr, resolution=10, boundary=0, limits=None):
        '''
        Initialise a distribution map. This is essentially a multidimensional 
        histogram produced from a set of data, arr(N,D) where N is the
        number of points and D is the number of dimensions. "resolution" 
        specifies the number of bins in each dimension, and may be a single 
        number or a list of length D. The optional limits parameter sets the
        histogram boundaries, which are otherwise set automatically to include
        all the data.
        '''
        self.ndim = arr.shape[1]
        self.boundary = boundary
        # resolution may be a single value or a list of values, 1 per dimension
        self.resolution = np.zeros(self.ndim)
        self.resolution[:] = resolution
        if limits is None:
            self.limits = []
            min = arr.min(axis=0)
            max = arr.max(axis=0)
            # set a boundary
            for i in range(self.ndim):
                buff = (max[i]-min[i])/(self.resolution[i]-2*boundary)*boundary*1.01
                self.limits.append((min[i]-buff,max[i]+buff))
        else:
            self.limits=limits

        # Create the histogram
        self._H, self._edges = np.histogramdd(arr, bins=self.resolution, range=self.limits)
        self.shape = self._H.shape
        # find the bin dimensions (this should be buff, but don't assume)
        self.cellsize = []
        for i in range(self.ndim):
            self.cellsize.append(self._edges[i][1]-self._edges[i][0])
        # calculate the bin volume, and number of sampled bins
        self.cellvol = self.cellsize[0]
        for l in self.cellsize[1:]:
            self.cellvol = self.cellvol*l
        self.coverage = self._H[np.where(self._H > 0)].size
        # correct coverage if there is a boundary:
        v0 = 1
        v1 = 1
        for d in self.resolution:
           v0 = v0*d
           v1 = v1*(d-2*self.boundary)
        self.deadspace = v0 - v1
        self.volume = v0
        # now give preliminary values to the cluster-related data.
        self.ID = self._H.copy()
        indMax = np.unravel_index(self.ID.argmax(),self.ID.shape)
        self.ID = np.where(self._H > 0, 1, 0)
        self.ID[indMax] = -1
        self.sizes = [0, arr.shape[0]]

    def map(self, vec):
        '''
        Returns the index of the bin that corresponds to point vec.
        '''
        if len(vec) != self.ndim:
            raise ValueError('Error - vector has wrong length.')
        indx = []
        i = 0
        for c in vec:
            indx.append(np.digitize((c,c),self._edges[i])[0]-1)
            i += 1
        return indx

    def unmap(self, indx):
        '''
        Returns a vector corresponding to the mid point of the
        bin with the given index.
        '''
        if len(indx) != self.ndim:
            raise ValueError('Error - index has wrong number of dimensions.')
        vec = []
        i = 0
        for c in indx:
            vec.append(self.limits[i][0]+c*self.cellsize[i]+ 0.5*self.cellsize[i])
            i += 1
        return vec


    def cluster_id(self, vec):
        '''
        Returns the cluster ID of the point vec.
        '''
        if len(vec) != self.ndim:
            raise ValueError('Error - vector has wrong length.')
        indx = []
        i = 0
        for c in vec:
            indx.append(np.digitize((c,c),self._edges[i])[0]-1)
            i += 1
        return self.ID[tuple(indx)]

    def cluster_centre(self, id):
        '''
        Returns a vector corresponding to the mid point of the
        cluster with the given id. It is assumed that these are
        the bins with negative cluster ids.
        '''
        if id < 0:
            id = -id
        w = np.where(self.ID == -id)
        e = []
        for i in range(self.ndim):
            e.append(self._edges[i][w[i][0]]+0.5*self.cellsize[i])
        return e 

    def cluster_size(self, id):
        '''
        Returns the number of samples in cluster = id
        '''
        return self.sizes[id-1]

def coco(map, method='coco', npoints=1):
    '''
    The CoCo (Complementary Coordinates) methods. The input is an
    N-dimensional histogram defining the occupancy of the space.
    Various CoCo methods will identify 'interesting' regions to
    be sampled next.
    '''
    if method == 'coco':
        '''
        returns new points, generated using the COCO procedure,
        in the form of an (npoints,D) numpy array, where D is the number of
        dimensions in the map.
        '''
        cp = np.zeros((npoints,map.ndim))
        # make a temporary binary image, and invert
        tmpimg = np.where(map._H > 0, 0, 1)
        for i in range(npoints):
            dis = nd.morphology.distance_transform_edt(tmpimg)
            indMax = np.unravel_index(dis.argmax(),dis.shape)
            for j in range(map.ndim):
                cp[i,j]=map._edges[j][0]+indMax[j]*map.cellsize[j]
            
            tmpimg[indMax] = 0
        return cp

    elif method == 'hpoints':
        '''
        hpoints returns new points that form a halo of unsampled space
        just beyond the sampled region.
        '''
        # This is the halo filter:
        def f(arr):
            cval = arr[len(arr)/2]
            if cval == 0 and np.max(arr) > 0:
                return 1
            else:
                return 0

        halo = nd.filters.generic_filter(map._H,f,size=3,mode='constant')
        npoints = int(np.sum(halo))
        hp = np.zeros((npoints,map.ndim))
        for i in range(npoints):
            indMax = np.unravel_index(halo.argmax(),map.shape)
            for j in range(map.ndim):
                hp[i,j]=map.edges[j][0]+indMax[j]*map.cellsize[j]
            
            halo[indMax] = 0
        return hp

    elif method == 'fpoints':
        '''
        fpoints returns new points at the frontier of sampled space
        '''
        # This is the frontier filter:
        def f(arr):
            cval = arr[len(arr)/2]
            if cval > 0 and np.min(arr) == 0:
                return 1
            else:
                return 0

        front = nd.filters.generic_filter(map._H,f,size=3,mode='constant')
        npoints = int(np.sum(front))
        fp = np.zeros((npoints,map.ndim))
        for i in range(npoints):
            indMax = np.unravel_index(front.argmax(),map.shape)
            for j in range(map.ndim):
                fp[i,j]=map._edges[j][0]+indMax[j]*map.cellsize[j]
            
            front[indMax] = 0
        return fp

    elif method == 'bpoints':
        '''
        bpoints() returns new points not at the frontier of sampled space
        '''
        # This is the buried filter:
        def f(arr):
            cval = arr[len(arr)/2]
            if cval > 0 and np.min(arr) > 0:
                return 1
            else:
                return 0

        bur = nd.filters.generic_filter(map._H,f,size=3,mode='constant')
        npoints = int(np.sum(bur))
        bp = np.zeros((npoints,map.ndim))
        for i in range(npoints):
            indMax = np.unravel_index(bur.argmax(),map.shape)
            for j in range(map.ndim):
                bp[i,j]=map._edges[j][0]+indMax[j]*map.cellsize[j]
            
            bur[indMax] = 0
        return bp

    elif method == 'rpoints':
        '''
        rpoints() returns one point per bin of sampled space, and its weight
        '''

        tmpimg = map._H.copy()
        hsum = np.sum(map._H)
        npoints = tmpimg[np.where(tmpimg > 0)].size
        wt = np.zeros((npoints))
        rp = np.zeros((npoints,map.ndim))
        for i in range(npoints):
            indMax = np.unravel_index(tmpimg.argmax(),map.shape)
            for j in range(map.ndim):
                rp[i,j]=map._edges[j][0]+indMax[j]*map.cellsize[j]
            
            tmpimg[indMax] = 0
            wt[i] = map._H[indMax]/hsum
        return rp,wt

    else:
        raise ValueError('Unknown method: {}'.format(method))




def watershed(map):
    '''
    Watershed clustering. The input is a map object that defines the
    sampling of the space. The method sets the ID data in the map.
    '''
    # The method requires a clear 1-bin boundary around the distribution:
    if map.boundary < 1:
        raise ValueError('The input map needs to have a boundary.')

    # define the function that will locate local maxima in the distibution: 
    def f(arr):
        if np.argmax(arr) == len(arr)/2:
            return 1
        else:
            return 0
    # now find them (calling them 'labels' but at the moment they are
    # really 'maxima', will become labels properly next):
    labels = nd.filters.generic_filter(map._H,f,size=3,mode='constant')
    # now turn them into proper labels:
    nd.measurements.label(labels,output=labels)
    # The array used for doing the watershedding needs to simultaneously
    # encode both the number of counts in the bin, and the bin label. We
    # do this by raising the bin counts by enough powers of ten to leave
    # room in the low order bits for the label.
    maxval = np.max(labels)
    scale = int(10**(np.ceil(np.log10(maxval+1))))
    
    # Now put all this into the ID array:
    map.ID = map.ID * scale + labels
    # Define the function that assigns bins to labels. For each currently
    # unlabelled bin, we search around to find the neighbour bin with
    # the highest occupancy. If this bin is already labelled, the current
    # bin gets this label too.
    def f2(arr,scale):
        cval = arr[len(arr)/2]
        if cval > 0 and cval%scale == 0:
            maxval = np.max(arr)
            if maxval%scale > 0:
                cval = cval +maxval%scale
        return cval

    # now we can do the watershed-like clustering. The process is iterative,
    # labels "spread out" in a downhill direction from the maxima until all
    # bins are labelled.
    IDnew = nd.filters.generic_filter(map.ID,f2,size=3,mode='constant',extra_arguments=(scale,))
    while np.any(IDnew != map.ID):
        map.ID = IDnew
        IDnew = nd.filters.generic_filter(map.ID,f2,size=3,mode='constant',extra_arguments=(scale,))

    # make ID nice for indexing:
    map.ID = map.ID.astype(np.int)
    # remove the bincount data, leaving just the labels:
    map.ID = map.ID%scale
    # Flatten for now to make the next few steps a bit easier:
    map.ID = map.ID.flatten()
    # now we find out big each cluster is:
    sizes = np.zeros(1+int(np.max(map.ID)), dtype=np.int)
    for i in range(len(map.ID)):
        sizes[int(map.ID[i])] += map._H.flatten()[i]
    # now reassign labels so largest group has label=1, etc.

    newlab = np.argsort(np.argsort(sizes))
    newlab = abs(newlab-newlab.max()) + 1
    map.sizes = np.sort(sizes)[::-1]

    for i in range(len(map.ID)):
        map.ID[i] = newlab[map.ID[i]]

    # put ID back into the right shape:
    map.ID = map.ID.reshape(map._H.shape)
    # now mark "root" structures with negative indices:
    map.ID = np.where(labels>0, -map.ID, map.ID)
    # set empty bins to have label=0:
    map.ID = np.where(map.ID==maxval+1, 0, map.ID)



#!/usr/bin/env python
"""  Setup script. Used by easy_install and pip. """

import os

from distutils.core import setup
from setuptools import find_packages

import sys
import textwrap

'''Discover the package version'''
import re
VERSIONFILE="pypcazip/pcazip/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RunTimeError("Unable to find version string in {}.".format(VERSIONFILE))

'''The minimum version of NumPy required.'''
min_numpy_version = '1.0.3'
min_scipy_version = '0.10.0'

'''Some functions for showing errors and warnings.'''
def _print_admonition(kind, head, body):
    tw = textwrap.TextWrapper(
        initial_indent='   ', subsequent_indent='   ')

    print ".. %s:: %s" % (kind.upper(), head)
    for line in tw.wrap(body):
        print line

def exit_with_error(head, body=''):
    _print_admonition('error', head, body)
    sys.exit(1)

def print_warning(head, body=''):
    _print_admonition('warning', head, body)

print "Checking python version!"
if not (sys.version_info[0] >= 2 and sys.version_info[1] >= 4):
    exit_with_error("You need Python 2.4 or greater to install pyPcazip!")
print "Python version checked."

''' Check for required Python packages. '''
def check_import(pkgname, pkgver):
    try:
        mod = __import__(pkgname)
    except ImportError:
        exit_with_error(
            "Can't find a local %s Python installation." % pkgname,
            "Please read carefully the ``README`` file "
            "and remember that pyPcazip needs the %s package "
            "to compile and run." % pkgname )
    else:
        if mod.__version__ < pkgver:
            exit_with_error(
                "You need %(pkgname)s %(pkgver)s or greater to run pyPcazip!"
                % {'pkgname': pkgname, 'pkgver': pkgver} )

    print ( "* Found %(pkgname)s %(pkgver)s package installed."
            % {'pkgname': pkgname, 'pkgver': mod.__version__} )
    globals()[pkgname] = mod
    
check_import('numpy', min_numpy_version)
check_import('scipy', min_scipy_version)

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup_args = {
    'name'             : "pyPcazip",
    'version'          : verstr,
    'description'      : "PCA-based trajectory file compression and analysis.",
    'long_description' : "PCA-based trajectory file compression and analysis.",
    'author'           : "The University of Nottingham & BSC",
    'url'              : "https://bitbucket.org/ramonbsc/pypcazip/overview",
    'download_url'     : "https://bitbucket.org/ramonbsc/pypcazip/get/"+verstr+".tar.gz",
    'license'          : "BSD license or similar.",
    'classifiers'      : [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: Python Software Foundation License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix'
    ],

    'packages'    : find_packages('pypcazip'),
    'package_dir' : {'': 'pypcazip'},
    'scripts' : ['pypcazip/pyPcazip',
                 'pypcazip/pyPcaunzip',
                 'pypcazip/pyPczcomp',
                 'pypcazip/pyPczdump',
                 'pypcazip/pyPczplot',
                 'pypcazip/pyPczclust'],
    'install_requires' : ['numpy',
                          'scipy',
    		          'MDAnalysis==0.11.0',
                          'argparse'],
    'extras_require': {'PCZ7': ['h5py', 'cython'],
                       'BINPOS': ['mdtraj', 'cython'],
                       'NETCDF': ['netCDF4', 'cython'],
                       'ALL': ['h5py',
                               'mdtraj', 'cython',
                               'netCDF4'],
                       },

    'zip_safe'         : False,
}

setup(**setup_args)



#!/usr/bin/env python -W ignore
'''
                 *** The command line interface for pyPczclust ***

                       Adapted to use the mapping module.
'''

import logging as log

import numpy as np
from scipy import ndimage

from MDPlus.analysis.pca import pczfile
from MDPlus.analysis import mapping

def pczclust(args): 
    '''
    Performs histogram/watershed based clustering on data from a .pcz
    file.
    '''

    if args.verbosity > 0:
        log.basicConfig(format="%(levelname)s: %(message)s", level=log.DEBUG)
        log.info("Verbose output")
    else:
        log.basicConfig(format="%(levelname)s: %(message)s")

    p = pczfile.Pczfile(args.pczfile)

    projs = np.zeros((p.nframes,args.dims))
    for i in range(args.dims):
        projs[:,i] = p.proj(i)

    m = mapping.Map(projs,resolution=args.bins, boundary=1)
    mapping.watershed(m)
    out = [m.cluster_id(id) for id in projs]

    np.savetxt(args.outfile,np.c_[projs,out], fmt=("%8.3f"*args.dims + "%5d"))
